import time
from typing import List, Optional

import requests
from joblib import Parallel, delayed, parallel_backend
from shopcloud_secrethub import SecretHub

from . import helpers


def github_fetch_pull_requests(owner: str, repo: str, **kwargs):
    if kwargs.get('is_simulate', False):
        return [{
            'merged': False,
            'number': 1,
            'mergeable': True,
            'title': '[SECURITY] Add automatic dependencies upgrade',
            'mergeable_state': 'clean',
        }]
    token = kwargs.get('api_token', None)
    if token is None:
        raise Exception('Missing API token')

    response = requests.get(
        f'https://api.github.com/repos/{owner}/{repo}/pulls',
        headers={
            'Authorization': f'token {token}',
        },
        params={
            'state': 'open',
            'sort': 'created',
            'direction': 'asc',
        },
    )

    if not (response.status_code >= 200 and response.status_code <= 299):
        raise Exception('Error while fetching pull requests')

    return response.json()


def github_fetch_pull_request(owner: str, repo: str, pull_request_number: int, **kwargs):
    if kwargs.get('is_simulate', False):
        return {
            'merged': False,
            'number': 1,
            'mergeable': True,
            'title': '[SECURITY] Add automatic dependencies upgrade',
        }
    token = kwargs.get('api_token', None)
    if token is None:
        raise Exception('Missing API token')

    response = requests.get(
        f'https://api.github.com/repos/{owner}/{repo}/pulls/{pull_request_number}',
        headers={
            'Authorization': f'token {token}',
        },
    )

    if not (response.status_code >= 200 and response.status_code <= 299):
        raise Exception('Error while fetching pull requests')

    return response.json()


def github_merge_pull_request(owner: str, repo: str, pull_request_number: int, **kwargs):
    if kwargs.get('is_simulate', False):
        return {}
    token = kwargs.get('api_token', None)
    if token is None:
        raise Exception('Missing API token')

    for i in range(10):
        response = requests.put(
            f'https://api.github.com/repos/{owner}/{repo}/pulls/{pull_request_number}/merge',
            headers={
                'Authorization': f'token {token}',
            },
        )

        print(response.text)
        if not (response.status_code >= 200 and response.status_code <= 299):
            if response.status_code == 403 and i < 9:
                time.sleep(30)
                continue
            raise Exception('Error while fetching pull requests')

    return response.json()


def get_api_token(**kwargs) -> Optional[str]:
    if kwargs.get('is_simulate', False):
        return 'test-token'
    hub = SecretHub(user_app="microservice-cli")
    return hub.read('talk-point/app-microservices-cli/production/github-key')


def run_repo_merge_security_pull_requests(owner: str, repo: str, **kwargs) -> Optional[bool]:
    try:
        is_simulate = kwargs.get('is_simulate', False)
        api_token = kwargs.get('api_token')

        pull_requests = [
            x for x in
            github_fetch_pull_requests(
                owner,
                repo,
                is_simulate=is_simulate,
                api_token=api_token,
            )
            if "[SECURITY] Add automatic dependencies upgrade" in x.get('title')
        ]
        if len(pull_requests) == 0:
            print(f'+ {owner}/{repo} - No pull requests found')
            return None

        pull_request = github_fetch_pull_request(
            owner,
            repo,
            pull_requests[0].get('number'),
            is_simulate=is_simulate,
            api_token=api_token
        )
        if pull_request.get('merged') in [None, True]:
            print(f'+ {owner}/{repo} - Already merged')
            return None

        if pull_request.get('mergeable') in [None, False]:
            url = pull_request.get('html_url')
            print(helpers.bcolors.FAIL + f'+ {owner}/{repo} - Can not be merged - {url}' + helpers.bcolors.ENDC)
            return None

        if pull_request.get('mergeable_state') != 'clean':
            url = pull_request.get('html_url')
            print(helpers.bcolors.FAIL + f'+ {owner}/{repo} - Can not be merged - {url}' + helpers.bcolors.ENDC)
            return None

        github_merge_pull_request(
            owner,
            repo,
            pull_request.get('number'),
            is_simulate=is_simulate,
            api_token=api_token
        )
        print(helpers.bcolors.OKGREEN + f'+ {owner}/{repo} - Merging pull request' + helpers.bcolors.ENDC)
        return True
    except Exception as e:
        print(helpers.bcolors.FAIL + f'+ {owner}/{repo} - Error while merging pull request - {e}' + helpers.bcolors.ENDC)
        return False


def load_repos(config) -> List[str]:
    repos = {x.get('repo') for x in config.load_projects() if x.get('repo') is not None}
    repos = [
        "/".join(x.replace('https://github.com/', '').split('/')[1:])
        for x in repos if x is not None
    ]
    return list(repos)


def cli_main(args, config):
    if args.action == 'merge-security-pull-requests':
        owner = 'Talk-Point'
        repos = [args.repo] if args.repo is not None else load_repos(config)
        api_token = get_api_token(is_simulate=args.simulate)

        with parallel_backend('threading', n_jobs=-1):
            Parallel()(
                delayed(run_repo_merge_security_pull_requests)(owner, m, is_simulate=args.simulate, api_token=api_token)
                for m in repos
            )
    elif args.action == 'ci-secrets-rolling':
        print('Got to https://github.com/organizations/Talk-Point/settings/secrets/actions')
        print('- PYPY_TOKEN - generate a new token on the PyPi website')
        print('- SECRETHUB_TOKEN - rotate the token for the github-ci user')
        print('- TOKEN - rotate the token for the github-ci user in TP-Server')

    return 0



import time
from typing import List, Optional

import requests
from shopcloud_secrethub import SecretHub

from . import helpers


def check_health(url: str, **kwargs) -> int:
    is_simulate = kwargs.get('is_simulate', False)
    status_code = None
    try:
        if is_simulate:
            return 200
        response = requests.get(url, timeout=30)
        return response.status_code
    except Exception:
        status_code = None
    return status_code



def load_base_urls(config) -> List[str]:
    items = {x.get('base_url') for x in config.load_projects() if x.get('base_url') is not None}
    return list(set(items))


def load_apps(config) -> List[dict]:
    return [x for x in config.load_projects() if x.get('name') is not None and "django" in x.get('types', [])]


def get_api_token(**kwargs) -> Optional[str]:
    if kwargs.get('is_simulate', False):
        return 'test-token'
    hub = SecretHub(user_app="microservice-cli")
    return hub.read('talk-point/app-microservices-cli/production/deploy-api-token')


def generate_release(data: dict, **kwargs) -> dict:
    try:
        is_simulate = kwargs.get('is_simulate', False)
        repo = data.get('repo')
        if repo is None:
            return {
                'status': 'ERROR',
                'message': 'Missing repo',
            }
        pieces = repo.replace('https://github.com/', '').split('/')
        if len(pieces) != 2:
            return {
                'status': 'ERROR',
                'message': 'Invalid repo',
            }
        owner = pieces[0]
        repo = pieces[1]
        if is_simulate:
            return {
                'status': 'OK',
            }
        for i in range(10):
            response = requests.post(
                'https://shopcloud-deploy-api-2yj8sc5z.ew.gateway.dev/pull-request/open',
                headers={
                    'X-Api-Key': get_api_token(is_simulate=is_simulate),
                },
                json={
                    'owner': owner,
                    'repo': repo,
                }
            )
            print(f'+ {repo} - {response.status_code}')
            if not (response.status_code >= 200 and response.status_code <= 299):
                print(f'+ {repo} - {response.text}')
                if response.status_code == 403 and i < 9:
                    time.sleep(30)
                    continue
                return {
                    'status': 'ERROR',
                    'message': 'Error while creating pull request',
                }

            return {
                'status': 'OK',
            }

        return {
            'status': 'OK',
        }
    except Exception as e:
        return {
            'status': 'ERROR',
            'message': str(e),
        }


def cli_main(args, config) -> int:
    return_code = 0
    if args.action == 'health':
        items = load_base_urls(config)
        datas = [{
            'url': x,
            'status_code': check_health(x, is_simulate=args.simulate)
        } for x in items]
        datas = [{
            **x,
            **{
                'status': 'DOWN' if x.get('status_code') is None else 'OK' if 200 <= x.get('status_code') < 300 else 'DOWN'
            }
        } for x in datas]

        if args.debug:
            print(datas)

        for d in datas:
            if d.get('status') == 'DOWN':
                print(f"- {d.get('url')} {helpers.bcolors.FAIL}DOWN{helpers.bcolors.ENDC}")
            elif d.get('status') == 'OK':
                print(f"- {d.get('url')} {helpers.bcolors.OKGREEN}{d.get('status_code')} OK{helpers.bcolors.ENDC}")

        if [x for x in datas if x.get('status') == 'DOWN']:
            return 1
    elif args.action == 'list':
        items = load_apps(config)
        for item in items:
            print(f"- {item.get('name')} {item.get('base_url')}")
    elif args.action == 'release':
        items = load_apps(config)
        if args.name is not None:
            items = [x for x in items if x.get('name') == args.name]
        items = [{
            'item': x,
            'action': generate_release(x, is_simulate=args.simulate),
        } for x in items]

        print('Result')

        for item in items:
            if item.get('action').get('status') == 'OK':
                print(helpers.bcolors.OKGREEN + f"- {item.get('item').get('name')} {item.get('action').get('status')}" + helpers.bcolors.ENDC)
            else:
                print(f"+ {item.get('item').get('name')} {item.get('action').get('status')} - {item.get('action').get('message')}")
                return_code = 1
    return return_code



import json
import os
import shlex
import subprocess
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Tuple

from joblib import Parallel, delayed

from . import helpers, steps
from .configs import Config


def database_list(project: str, instance_id: str, simulate: bool = False) -> List[str]:
    command = f'gcloud sql databases list --project="{project}" --instance="{instance_id}" --format=json'
    if simulate:
        data = '[{"name": "mysql", "name": "db-a"}]'
    else:
        p = subprocess.run(shlex.split(command), stdout=subprocess.PIPE)
        data = p.stdout.decode("utf-8")

    try:
        data = json.loads(data)
    except Exception:
        print(helpers.bcolors.FAIL + f"[project/{project}/instance/{instance_id}] Error while parsing json" + helpers.bcolors.ENDC)
        return []

    data = [x.get("name") for x in data]
    data = [
        x
        for x in data
        if x not in ["mysql", "information_schema", "performance_schema", None, "sys"]
    ]

    return data


def unpack_db_name(name: Optional[str]) -> Tuple[bool, str, str, str, str]:
    if name is None:
        print(
            helpers.bcolors.FAIL
            + "Insert a SQL Instanz connection string like <project>:<region>:<instance-id>"
            + helpers.bcolors.ENDC
        )
        return False, None, None, None, None

    values = name.split(":")
    if len(values) not in [3, 4]:
        print(
            helpers.bcolors.FAIL
            + "Insert a SQL Instanz connection string like <project>:<region>:<instance-id>"
            + helpers.bcolors.ENDC
        )
        return False, None, None, None, None

    project = values[0]
    region = values[1]
    instance_id = values[2]
    try:
        db = values[3]
    except IndexError:
        db = None
    return True, project, region, instance_id, db


def instances_list(config) -> List[str]:
    databases = [x.get("databases") for x in config.load_projects()]
    databases = [x for x in databases if x is not None]
    databases = [item for sublist in databases for item in sublist]
    databases = [x for x in databases if x.strip().startswith("sql:")]
    databases = [x for x in databases if len(x.split(":")) == 5]
    databases = {":".join(x.split(":")[1:-1]) for x in databases}
    return databases


def export_database(
    config: Config, project: str, instance_id: str, db: str, hash: Optional[str], simulate: bool
) -> int:
    if hash is None:
        hash = helpers.random_string(5)
    now = datetime.now()
    bucket_name = f"{instance_id.lower()}-db-exports"
    backup_name = f"{now.strftime('%Y')}/{now.strftime('%m')}/{now.strftime('%d')}/{hash}/{project}-{instance_id}-{db}.sql"
    cmd = f"gcloud sql export sql {instance_id} gs://{bucket_name}/{backup_name} --database='{db}' --project='{project}'"
    manager = steps.Manager(
        config,
        [steps.StepCommand(shlex.split(cmd))],
        simulate=simulate,
    )
    rc = manager.run()
    return rc


def cli_main(args, config) -> int:
    if args.action == "sql-init":
        is_success, project, region, instance_id, _ = unpack_db_name(args.name)
        if not is_success:
            return 1

        bucket_name = f"{instance_id.lower()}-db-exports"

        print(f"- Bucket gs://{bucket_name} wird angelegt")
        manager = steps.Manager(
            config,
            [
                steps.StepCommand(
                    shlex.split(
                        f"gcloud storage buckets create gs://{bucket_name} --project='{project}' --location='{region}'"
                    ),
                    can_fail=True,
                    retry=0,
                ),
            ],
            simulate=args.simulate,
        )
        rc = manager.run()
        if rc != 0:
            return rc

        print(
            "- dem Dienstkonto des SQL Servers Administrator Rechte für Storage im Bucket geben"
        )
        print(
            "- im Bucket den Lebenszyklus einstellen, das die Dateien nach 14 Tagen gelöscht werden"
        )
        print(
            f"https://console.cloud.google.com/storage/browser/{bucket_name};tab=lifecycle?project={project}"
        )

    elif args.action == "sql-list-instances":
        instances = instances_list(config)

        for instance in instances:
            print(f"- {instance}")

    elif args.action == "sql-list-databases":
        is_success, project, region, instance_id, _ = unpack_db_name(args.name)
        if not is_success:
            return 1

        databases = database_list(project, instance_id, args.simulate)
        print(databases)
    elif args.action == "sql-dump":
        is_success = True
        datas = []
        if args.name is None:
            instances = instances_list(config)
            datas = [
                x.get('databases') for x in config.load_projects() if x.get("databases") is not None
            ]
            datas = [item for sublist in datas for item in sublist]
            datas = [x[4:] for x in datas if x.strip().startswith("sql:")]
            datas = [unpack_db_name(x) for x in datas]
            datas = {
                f'project/{project}/locations/{region}/instances/{instance_id}':
                {
                    'project': project,
                    'region': region,
                    'instance_id': instance_id,
                    'databases': []
                }
             for _, project, region, instance_id, _ in datas}
            datas = [datas[x] for x in datas.keys()]
            datas = [{
                **x,
                'databases': database_list(x['project'], x['instance_id'], args.simulate)
            } for x in datas]
            if len([x for x in datas if len(x['databases']) <= 0]):
                is_success = False # kompletter abbruch eine instance konnten die dbs nicht geladen werden
        else:
            is_success, project, region, instance_id, db = unpack_db_name(args.name)
            if not is_success:
                return 1
            if db is None:
                databases = database_list(project, instance_id, args.simulate)
            else:
                databases = [db]
            datas = [{
                'project': project,
                'region': region,
                'instance_id': instance_id,
                'databases': databases,
            }]

        # liste instanzen mit liste der dbs
        def run_export(config, instance: dict, hash: str, is_simulate=False):
            databases = [
                {
                    'name': x,
                    'status': export_database(config, instance.get('project'), instance.get('instance_id'), x, hash, is_simulate)
                }
                for x in instance.get('databases', [])
            ]
            return {
                **instance,
                **{
                    'databases': databases,
                }
            }

        hash = helpers.random_string(5)
        datas = Parallel(backend="threading" ,n_jobs=-1)(delayed(run_export)(config, x, hash, args.simulate) for x in datas)

        print('Results:')
        for instance in datas:
            if len(instance.get('databases', [])) <= 0:
                is_success = False
                print(
                    helpers.bcolors.FAIL
                    + f"- {instance.get('project')}/{instance.get('instance_id')} - no databases found"
                    + helpers.bcolors.ENDC
                )
                continue

            print(f"- {instance.get('project')}/{instance.get('instance_id')}")
            for db in instance.get('databases', []):
                if db.get('status') != 0:
                    is_success = False
                    print(
                        helpers.bcolors.FAIL
                        + f"  - {db.get('name')}"
                        + helpers.bcolors.ENDC
                    )
                else:
                    print(
                        helpers.bcolors.OKGREEN
                        + f"  - {db.get('name')}"
                        + helpers.bcolors.ENDC
                    )


        if not is_success: # check global is ok
            return 1

        return 0
    elif args.action == "sql-download":
        instances = instances_list(config)
        home_dir = os.path.expanduser("~")

        def download_instances_backups(config: Config, instance: str) -> int:
            is_success, project, region, instance_id, _ = unpack_db_name(instance)
            if not is_success:
                return 1
            bucket_name = f"{instance_id.lower()}-db-exports"
            backup_path = f"{home_dir}/db_dumps/{bucket_name}"
            manager = steps.Manager(
                config,
                [
                    steps.StepCommand(
                        shlex.split(f"mkdir -p {backup_path}"),
                        can_fail=True,
                        retry=0,
                    ),
                    steps.StepCommand(
                        shlex.split(
                            f"gsutil rsync -r -u gs://{bucket_name}/ {backup_path}"
                        ),
                        can_fail=True,
                        retry=0,
                    ),
                ],
                simulate=args.simulate,
            )
            rc = manager.run()
            return rc
        res = Parallel(backend="threading",n_jobs=-1)(delayed(download_instances_backups)(config, instance) for instance in instances)
        if any(x != 0 for x in res):
            print("Ein Export ist fehlerhaft gelaufen.")
    elif args.action == "sql-push-to-drive":
        instances = instances_list(config)
        home_dir = os.path.expanduser("~")

        destination = args.name
        destination_path = Path(destination)
        if not destination_path.exists():
            print(
                helpers.bcolors.FAIL
                + f"Path not exists: {destination}"
                + helpers.bcolors.ENDC
            )

        manager = steps.Manager(
            config,
            [
                steps.StepCommand(
                    shlex.split(
                        f"rsync -av --ignore-existing {home_dir}/db_dumps/ {destination}"
                    ),
                    can_fail=True,
                    retry=0,
                ),
            ],
            simulate=args.simulate,
        )
        rc = manager.run()
        if rc != 0:
            return rc

    return 0
